package irnatura.offline;

import irnatura.offline.BgFrame;
import irnatura.offline.ImageWrap;
import irnatura.offline.TempProfile;

public class BgByte extends BgFrame
{
    protected final byte MAX = 127;
    protected final byte MIN = -127;
    protected final byte OVF = -128;

    public BgByte()
    {
        m_bytes = null;
        m_origin = 0;
        m_num_samples = 0;
        m_overflow = null;
        m_overflow_size = 0;
    }

    // Copy constructor -- really for BgByteIterator
    protected BgByte(BgByte bb)
    {
        m_bytes = bb.m_bytes;
        m_origin = bb.m_origin;
        m_num_samples = bb.m_num_samples;
        m_overflow = bb.m_overflow;
        m_overflow_size = bb.m_overflow_size;
    }

    protected byte [] m_bytes; // The deltas in [MIN,MAX], or OVF in m_overflow.

    // NOTE: should be same as getSize()
    public int length() { return (m_bytes == null) ? 0 : m_bytes.length; }

    protected short [] m_overflow; // Array of deltas too big to fit in a byte
    protected int m_overflow_size;

    private void overflow_reserve(int size) {
        if (m_overflow == null) {
            m_overflow = new short[32];
        }
        else if (size >= m_overflow.length) {
            short [] old = m_overflow;
            int newlength = 2* m_overflow.length;
            m_overflow = new short[newlength];
            for (int i = 0; i < old.length; ++i) {
                m_overflow[i] = old[i];
            }
        }
    }

    private void overflow_push(short val)
    {
        ++m_overflow_size;
        overflow_reserve(m_overflow_size);
        m_overflow[m_overflow_size] = val;
    }

    private void overflow_clear() { m_overflow_size = 0; }

    // Callback for load
    protected void initStorage(int size)
    {
        if (m_bytes == null || m_bytes.length != size)
            m_bytes = new byte[size];
        overflow_clear();
    }

    // Callback for load
    protected void setDelta(int index, short delta)
    {
        if (delta >= MIN && delta <= MAX) {
            m_bytes[index] = (byte)delta;
        }
        else {
            m_bytes[index] = OVF;
            overflow_push(delta);
        }
    }

    // Return number of discarded data points (from old or new).
    public int add(ImageWrap img)
    {
        // TODO: this invalidates m_badbits: reset?
        //       or pass mode/threshold to applyTo?
        if (m_bytes == null) {
            load(img);
            return 0;
        }
        int num_discarded = 0;
        if (m_bytes.length != img.getSize()) {
            throw new java.lang.ArrayIndexOutOfBoundsException(
                    "Bg length=" + m_bytes.length +
                    " img size=" + img.getSize());
        }

        int overflow_counter = 0;
        for (int i = 0; i < m_bytes.length; ++i) {
            short newdelta = (short)(img.getTemp(i) - m_origin);
            if (m_bytes[i] == OVF) {
                short curdelta = m_overflow[overflow_counter];
                short avgdelta = runningAvg(curdelta, newdelta);
                m_overflow[overflow_counter] = avgdelta;
                ++overflow_counter;
            }
            else {
                short curdelta = m_bytes[i];
                short avgdelta = runningAvg(curdelta, newdelta);
                if (avgdelta >= MIN && avgdelta <= MAX)
                    m_bytes[i] = (byte)avgdelta;
                else
                    ++num_discarded;
            }
        }
        ++m_num_samples;
        return num_discarded;
    }

    public String getDescription()
    {
        return super.getDescription() + "overflow=" + m_overflow_size + ";";
    }

    public TempForwardIterator iterator()
    {
        return new BgByteIterator(this);
    }
}

class BgByteIterator extends BgByte implements TempForwardIterator
{
    private int m_overflow_counter;
    private int m_last_request; // sanity

    public BgByteIterator(BgByte bb)
    {
        super(bb);
        m_overflow_counter = 0;
        m_last_request = 0;
    }

    public int getDelta(int i)
    {
        if (m_last_request >= i)                                     // sanity
            throw new java.lang.ArithmeticException("Out of order"); // sanity
        m_last_request = i;                                          // sanity

        short delta;
        if (m_bytes[i] == OVF) {
            delta = m_overflow[m_overflow_counter];
            ++m_overflow_counter;
        }
        else {
            delta = m_bytes[i];
        }
        return delta;
    }

    // FIXME: row/column major is confused
    public int getTemp(int x, int y)
    {
        return m_origin + getDelta(x + y * m_width);
    }

    public int getTemp(int i)
    {
        return m_origin + getDelta(i);
    }
}
