package irnatura.offline;

import irnatura.offline.ImageWrap;
import java.util.TreeSet;

interface BadBitsManager
{
    public boolean isBad(int index, int temp);
    public void setThreshold(float threshold);
    public TreeSet<Integer> getIndices(BgFrame bg);
    public int countBadBits(BgFrame bg);
}

class ImgDevBadBits implements BadBitsManager
{
    private float m_threshold;
    private int m_mean;
    private int m_stdev;
    private int m_threshtemp; // = m_stdev * m_threshold

    ImgDevBadBits(float threshold, BgFrame bg)
    {
        TempProfile tp = new TempProfile(bg.iterator());
        m_mean = tp.getMean();
        m_stdev = tp.getStdev();
        setThreshold(threshold);
    }

    public boolean isBad(int index, int temp)
    {
        return (Math.abs(temp) > m_threshtemp);
    }

    public TreeSet<Integer> getIndices(BgFrame bg)
    {
        TreeSet<Integer> indices = new TreeSet<Integer>();
        TempForwardIterator it = bg.iterator();
        for (int i = 0; i < it.getSize(); ++i) {
            int delta = it.getDelta(i);
            if (isBad(i, delta)) {
                indices.add(i);
            }
        }
        return indices;
    }

    // Note: linear time
    public int countBadBits(BgFrame bg)
    {
        TempForwardIterator it = bg.iterator();
        int numbad = 0;
        for (int i = 0; i < it.getSize(); ++i) {
            int delta = it.getDelta(i);
            if (isBad(i, delta)) {
                ++numbad;
            }
        }
        return numbad;
    }

    public void setThreshold(float threshold)
    {
        m_threshold = threshold;
        m_threshtemp = Math.round(m_stdev * m_threshold);
    }
}

public abstract class BgFrame
{
    public static final int MODE_NULL = 0;
    public static final int MODE_AVG = 1; // compute mean
    public static final int MODE_NEIGHBOR = 2; // dev wrt neighbors
    public static final int MODE_IMGDEV = 3; // stdev wrt frame
    public static final int MODE_PIXDEV = 4; // stdev per pixel

    private static final String [] MODESTRS = {
        "null",
        "avg",
        "neighbor",
        "imgdev",
        "pixdev",
    };

    public static String modestr(int m)
    {
        return (m >=0 && m < MODESTRS.length) ? MODESTRS[m] : "*invalid*";
    }

    //----------------------------------
    // Badbit stuff

    protected int m_mode; // A MODE_* value.
    protected BadBitsManager m_badbits;

    // Return whether changed
    public void setBadBitMode(int mode, float threshold)
    {
        if (m_mode == mode) {
            if (m_badbits != null) m_badbits.setThreshold(threshold);
            return;
        }
        m_mode = mode;
        switch (mode) {
            case MODE_IMGDEV:
                m_badbits = new ImgDevBadBits(threshold, this);
                return;
            case MODE_NULL:
            case MODE_AVG:
            case MODE_NEIGHBOR:
            case MODE_PIXDEV:
            default:
                m_badbits = null;
                return;
        }
    }

    //----------------------------------
    // Common Data and Methods

    protected int m_origin; // The temp against which deltas are based.
    protected int m_num_samples; // Number of sample frames collected.

    public boolean isLoaded() { return m_num_samples > 0; }

    public void applyTo(ImageWrap img)
    {
        TempForwardIterator it = iterator();
        for (int i = 0; i < img.getSize(); ++i) {
            int delta = it.getDelta(i);
            if (m_badbits != null && m_badbits.isBad(i, delta))
                continue;
            int newval = img.getTemp(i) - delta;
            img.setTemp(i, newval);
        }
    }

    public TreeSet<Integer> getBadBitIndices()
    {
        return (m_badbits == null) ? null : m_badbits.getIndices(this);
    }

    // Returns the number of pixels more than 8 bits
    // Threshold is for MODE_IMGDEV (and ...)
    public int load(ImageWrap img)
    {
        m_width = img.getWidth();
        m_height = img.getHeight();
        m_mode = 0;
        m_badbits = null;

        TempProfile tp = img.getOrComputeTempProfile();
        m_origin = tp.getMean();
        if (m_origin == TempProfile.INVALID)
            throw new java.lang.ArithmeticException("Invalid mean");
        initStorage(img.getSize());

        int num_big = 0;
        for (int i = 0; i < img.getSize(); ++i) {
            short delta = (short)(img.getTemp(i) - m_origin);
            setDelta(i, delta);
            if (delta < -128 || delta > 127)
                ++num_big;
        }
        ++m_num_samples;
        return num_big;
    }

    //----------------------------------
    // Helpers

    protected short runningAvg(short curval, short newval)
    {
        return (short)Math.round(((float)((curval * m_num_samples) + newval))
                                / (float)(m_num_samples + 1));
    }

    //--------------------------------------
    // TempForwardIterator partial interface

    protected int m_width;
    protected int m_height;

    public int getWidth() { return m_width; }
    public int getHeight() { return m_height; }
    public int getSize() { return m_width * m_height; }

    public String getDescription()
    {
        return "mode=" + modestr(m_mode) + ";"
             + "origin=" + m_origin + ";"
             + "samples=" + m_num_samples + ";"
             + ((m_badbits != null)
                    ? ("badbits=" + m_badbits.countBadBits(this) + ";") : "")
             ;
    }

    //----------------------------------
    // Abstract interface

    public abstract int length();

    // Return number of discarded data points (from old or new).
    public abstract int add(ImageWrap img);

    public abstract TempForwardIterator iterator();

    // Callback for load
    protected abstract void initStorage(int size);
    protected abstract void setDelta(int index, short delta);
}
